/*
** File: main.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent entry point into ADMXRC3 module driver
**
** (C) Copyright Alpha Data 2009-2012
*/

#include "model_boot.h"
#include "clock.h"
#include "device.h"
#include "derror.h"
#include "diagnostic.h"
#include "dma.h"
#include "dmabus.h"
#include "event.h"
#include "flash.h"
#include "fpga.h"
#include "identify.h"
#include "info.h"
#include "lock.h"
#include "main.h"
#include "module.h"
#include "pio.h"
#include "vpd.h"

#if !ADB3_MONOLITHIC
const char g_dfDriverName[] = "admxrc3";
#endif

/* User-mode interface for ADMXRC3_* API functions */
static const char g_interfaceClassName[] = "admxrc3";
static DfInterfaceClass* g_pInterfaceClass = NULL;
static DfInterfaceMethods g_interfaceMethods;

static DfIoStatus
onIoctl(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfClientRequest* pReq,
  unsigned int ioctlCode,
  void* pIoctl,
  unsigned int inSize,
  unsigned int outSize)
{
  DfIoStatus status = DfIoStatusSuccess;
  Admxrc3ClientContext* pClCtx = (Admxrc3ClientContext*)dfClientGetContext(pClObj);
  Admxrc3DeviceContext* pDevCtx = (Admxrc3DeviceContext*)pClCtx->pDevCtx;

  dfDebugPrint(3, ("onIoctl(admxrc3): entered pDevObj=%p pClObj=%p pReq=%p ioctlCode=0x%lx pIoctl=%p inSize=0x%lx outSize=0x%lx\n",
    pDevObj, pClObj, pReq, (unsigned long)ioctlCode, pIoctl, (unsigned long)inSize, (unsigned long)outSize));

  switch (ioctlCode) {
  case ADMXRC3_IOCTLCODE_CONFIGURE:
    status = ioctlConfigure(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_ERASEFLASH:
    status = ioctlEraseFlash(pDevCtx, pClCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETBANKINFO:
    status = ioctlGetBankInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCARDINFO:
    status = ioctlGetCardInfo(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCLOCKFREQUENCY:
    status = ioctlGetClockFrequency(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCOMMONBUFFERCOUNT:
    status = ioctlGetCommonBufferCount(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCOMMONBUFFERINFO:
    status = ioctlGetCommonBufferInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETDRIVERVERSION:
    status = ioctlGetDriverVersion(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETFPGAINFO:
    status = ioctlGetFpgaInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETMODULEINFO:
    status = ioctlGetModuleInfo(pDevCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_GETWINDOWINFO:
    status = ioctlGetWindowInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETFLASHBLOCKINFO:
    status = ioctlGetFlashBlockInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETFLASHINFO:
    status = ioctlGetFlashInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_LOCK:
    status = ioctlLock(pDevCtx, pClCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_NOTIFICATIONWAIT:
    status = ioctlNotificationWait(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READ:
    status = ioctlRead(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READDMA:
    status = ioctlReadDma(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READDMABUS:
    status = ioctlReadDmaBus(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READDMALOCKED:
    status = ioctlReadDmaLocked(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READFLASH:
    status = ioctlReadFlash(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READVPD:
    status = ioctlReadVpd(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

#if DF_HAS_USER_EVENT_HANDLE
  case ADMXRC3_IOCTLCODE_REGISTEREVENT:
    status = ioctlRegisterEvent(pDevCtx, pClCtx, pIoctl, inSize);
    break;
#endif

  case ADMXRC3_IOCTLCODE_SETCLOCKFREQUENCY:
    status = ioctlSetClockFrequency(pDevCtx, pClCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_SYNCFLASH:
    status = ioctlSyncFlash(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_UNCONFIGURE:
    status = ioctlUnconfigure(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_UNLOCK:
    status = ioctlUnlock(pDevCtx, pClCtx, pIoctl, inSize);
    break;

#if DF_HAS_USER_EVENT_HANDLE
  case ADMXRC3_IOCTLCODE_UNREGISTEREVENT:
    status = ioctlUnregisterEvent(pDevCtx, pClCtx, pIoctl, inSize);
    break;
#endif

  case ADMXRC3_IOCTLCODE_WRITE:
    status = ioctlWrite(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEDMA:
    status = ioctlWriteDma(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEDMABUS:
    status = ioctlWriteDmaBus(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEDMALOCKED:
    status = ioctlWriteDmaLocked(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEFLASH:
    status = ioctlWriteFlash(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEVPD:
    status = ioctlWriteVpd(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCARDINFOEX:
    status = ioctlGetCardInfoEx(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETSENSORINFO:
    status = ioctlDiagGetSensorInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_READSENSOR:
    status = ioctlDiagReadSensor(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETDEVICESTATUS:
    status = ioctlGetDeviceStatus(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_CLEARDEVICEERRORS:
    status = ioctlClearDeviceErrors(pDevCtx, pClCtx);
    break;

  default:
    status = DfIoStatusInvalidCode;
    break;
  }

  if (DfIoStatusDeferred != status) {
    dfRequestComplete(pReq, status);
  }
  return status;
}

#if DF_NEED_THUNK
static DfIoStatus
onIoctlThunk(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfClientRequest* pReq,
  unsigned int ioctlCode,
  void* pIoctl,
  unsigned int inSize,
  unsigned int outSize)
{
  DfIoStatus status = DfIoStatusSuccess;
  Admxrc3ClientContext* pClCtx = (Admxrc3ClientContext*)dfClientGetContext(pClObj);
  Admxrc3DeviceContext* pDevCtx = (Admxrc3DeviceContext*)pClCtx->pDevCtx;

  dfDebugPrint(3, ("onIoctlThunk: entered pDevObj=%p pClObj=%p pReq=%p ioctlCode=0x%lx pIoctl=%p inSize=0x%lx outSize=0x%lx)\n",
    pDevObj, pClObj, pReq, (unsigned long)ioctlCode, pIoctl, (unsigned long)inSize, (unsigned long)outSize));

  switch (ioctlCode) {
  case ADMXRC3_IOCTLCODE_CONFIGURE:
    status = ioctlConfigureThunk(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_ERASEFLASH:
    status = ioctlEraseFlash(pDevCtx, pClCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETBANKINFO:
    status = ioctlGetBankInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCARDINFO:
    status = ioctlGetCardInfo(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCLOCKFREQUENCY:
    status = ioctlGetClockFrequency(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCOMMONBUFFERCOUNT:
    status = ioctlGetCommonBufferCount(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCOMMONBUFFERINFO:
    status = ioctlGetCommonBufferInfoThunk(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETDRIVERVERSION:
    status = ioctlGetDriverVersion(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETFLASHINFO:
    status = ioctlGetFlashInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETFLASHBLOCKINFO:
    status = ioctlGetFlashBlockInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETFPGAINFO:
    status = ioctlGetFpgaInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETMODULEINFO:
    status = ioctlGetModuleInfoThunk(pDevCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_GETWINDOWINFO:
    status = ioctlGetWindowInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_LOCK:
    status = ioctlLockThunk(pDevCtx, pClCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_NOTIFICATIONWAIT:
    status = ioctlNotificationWait(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READ:
    status = ioctlReadThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READDMA:
    status = ioctlReadDmaThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READDMABUS:
    status = ioctlReadDmaBusThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READDMALOCKED:
    status = ioctlReadDmaLockedThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READFLASH:
    status = ioctlReadFlashThunk(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_READVPD:
    status = ioctlReadVpdThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

# if DF_HAS_USER_EVENT_HANDLE
  case ADMXRC3_IOCTLCODE_REGISTEREVENT:
    status = ioctlRegisterEventThunk(pDevCtx, pClCtx, pIoctl, inSize);
    break;
# endif

  case ADMXRC3_IOCTLCODE_SETCLOCKFREQUENCY:
    status = ioctlSetClockFrequency(pDevCtx, pClCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_SYNCFLASH:
    status = ioctlSyncFlash(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_UNCONFIGURE:
    status = ioctlUnconfigure(pDevCtx, pClCtx, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_UNLOCK:
    status = ioctlUnlock(pDevCtx, pClCtx, pIoctl, inSize);
    break;

# if DF_HAS_USER_EVENT_HANDLE
  case ADMXRC3_IOCTLCODE_UNREGISTEREVENT:
    status = ioctlUnregisterEventThunk(pDevCtx, pClCtx, pIoctl, inSize);
    break;
# endif

  case ADMXRC3_IOCTLCODE_WRITEDMA:
    status = ioctlWriteDmaThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEDMABUS:
    status = ioctlWriteDmaBusThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEDMALOCKED:
    status = ioctlWriteDmaLockedThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEFLASH:
    status = ioctlWriteFlashThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITE:
    status = ioctlWriteThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_WRITEVPD:
    status = ioctlWriteVpdThunk(pDevCtx, pClCtx, pReq, pIoctl, inSize);
    break;

  case ADMXRC3_IOCTLCODE_GETCARDINFOEX:
    status = ioctlGetCardInfoEx(pDevCtx, pIoctl, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETSENSORINFO:
    status = ioctlDiagGetSensorInfo(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_READSENSOR:
    status = ioctlDiagReadSensor(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_GETDEVICESTATUS:
    status = ioctlGetDeviceStatus(pDevCtx, pIoctl, inSize, outSize);
    break;

  case ADMXRC3_IOCTLCODE_CLEARDEVICEERRORS:
    status = ioctlClearDeviceErrors(pDevCtx, pClCtx);
    break;

  default:
    status = DfIoStatusInvalidCode;
    break;
  }

  if (DfIoStatusDeferred != status) {
    dfRequestComplete(pReq, status);
  }
  return status;
}
#endif

static DfIoStatus
onOpen(
	DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfClientRequest* pReq,
  DfInterface* pInterface,
  boolean_t bPrivileged)
{
  Admxrc3DeviceContext* pDevCtx;
  Admxrc3ClientContext* pClCtx = NULL;
  unsigned int i;
  DfSpinLockFlags f;
  DfIoStatus status = DfIoStatusSuccess;

  pDevCtx = (Admxrc3DeviceContext*)dfInterfaceGetContext(pInterface);

  pClCtx = (Admxrc3ClientContext*)dfMalloc(sizeof(Admxrc3ClientContext));
  if (NULL == pClCtx) {
    status = DfIoStatusNoMemory;
    goto done;
  }
  
  pClCtx->pDevCtx = pDevCtx;
  pClCtx->bPrivileged = bPrivileged;
  for (i = 0; i < pDevCtx->info.numFlashBank; i++) {
    pClCtx->bNeedFlashSync[i] = FALSE;
  }
  pClCtx->bNeedVpdSync = FALSE;

#if !defined(ADMXRC3_LOCKED_BUFFERS_GLOBAL) && ADMXRC3_LOCKED_BUFFERS_GLOBAL
  /* Initialize DMA buffer stuff */
  dfSpinLockInit(&pClCtx->lockedBuffers.lock);
  pClCtx->lockedBuffers.stack.top = 0;
  for (i = 0; i < MAX_NUM_LOCKED_USER_SPACE_BUFFER; i++) {
    /* Handle actually returned to user-space app is 'i + 1' */
    pClCtx->lockedBuffers.stack.free[i] = i;
  }
#endif

  pClCtx->hDisableSleep = dfSystemSleepDisable(pDevCtx->pDevObj);
  if (NULL == pClCtx->hDisableSleep) {
    status = DfIoStatusNoMemory;
    goto done;
  }

  f = dfSpinLockGet(&pDevCtx->powerMgmt.lock);
  if (pDevCtx->powerMgmt.bDeferOpen) {
    dfDebugPrint(1, ("onOpen: deferring open\n"));
    status = DfIoStatusDeferred;
    dfRequestPending(pReq);
    pClCtx->pOpenRequest = pReq;
    dfListAddToTail(&pDevCtx->powerMgmt.waitingOpen, &pClCtx->waitNode);
    dfSpinLockPut(&pDevCtx->powerMgmt.lock, f);
  } else {
    pDevCtx->powerMgmt.openCount++;
    dfSpinLockPut(&pDevCtx->powerMgmt.lock, f);
  }

done:
  if (status != DfIoStatusDeferred) {
    if (DfIoStatusIsError(status)) {
      if (NULL != pClCtx->hDisableSleep) {
        dfSystemSleepReenable(pDevCtx->pDevObj, pClCtx->hDisableSleep);
      }
      if (NULL != pClCtx) {
        dfFree(pClCtx);
      }
    } else {
      dfClientSetContext(pClObj, pClCtx);
    }
    dfRequestComplete(pReq, status);
  }
  return status;
}

static void
onCleanup(
	DfDeviceObject* pDevObj,
  DfClientObject* pClObj)
{
  Admxrc3ClientContext* pClCtx = (Admxrc3ClientContext*)dfClientGetContext(pClObj);
  Admxrc3DeviceContext* pDevCtx = (Admxrc3DeviceContext*)pClCtx->pDevCtx;
  unsigned int i;
  DfSpinLockFlags f;

  /* Clean up DMA transfers, i.e. cancel any that are still ongoing */
  cleanupDma(pDevCtx, pClCtx);

  /* If this client still owns any target FPGAs, release ownership */
  for (i = 0; i < pDevCtx->info.numTargetFpga; i++) {
    f = dfSpinLockGet(&pDevCtx->fpga[i].owner.lock);
    if (pClCtx == pDevCtx->fpga[i].owner.pClCtx) {
      pDevCtx->fpga[i].owner.pClCtx = NULL;
    }
    dfSpinLockPut(&pDevCtx->fpga[i].owner.lock, f);
  }
}

static void
onClose(
	DfDeviceObject* pDevObj,
  DfClientObject* pClObj)
{
  Admxrc3ClientContext* pClCtx = (Admxrc3ClientContext*)dfClientGetContext(pClObj);
  Admxrc3DeviceContext* pDevCtx = (Admxrc3DeviceContext*)pClCtx->pDevCtx;
  unsigned int i;
  DfSpinLockFlags f;

  for (i = 0; i < pDevCtx->info.numFlashBank; i++) {
    if (pClCtx->bNeedFlashSync[i]) {
      syncFlashOnClose(pDevCtx, pClCtx, i);
      pClCtx->bNeedFlashSync[i] = FALSE;
    }
  }
  if (pClCtx->bNeedVpdSync) {
    syncVpdOnClose(pDevCtx, pClCtx);
    pClCtx->bNeedVpdSync = FALSE;
  }

#if DF_HAS_USER_EVENT_HANDLE
  /* Clean up lists of registered user events */
  cleanupUserEvents(pDevCtx, pClCtx);
#endif

  /* Clean up locked user-space buffers */
  cleanupLockedBuffers(pDevCtx, pClCtx);

  f = dfSpinLockGet(&pDevCtx->powerMgmt.lock);
  pDevCtx->powerMgmt.openCount--;
  dfSpinLockPut(&pDevCtx->powerMgmt.lock, f);

  dfSystemSleepReenable(pDevCtx->pDevObj, pClCtx->hDisableSleep);

  /* Free the memory associated with the client's context */
  if (NULL != pClCtx) {
    dfFree(pClCtx);
    dfClientSetContext(pClObj, NULL);
  }
}

static void
freeDeviceContext(
	Admxrc3DeviceContext* pDevCtx)
{
  if (pDevCtx->userMode.bRegistered) {
    if (!dfInterfaceUnregister(pDevCtx->pDevObj, &pDevCtx->userMode.iface)) {
      dfDebugPrint(0, ("*** freeDeviceContext: failed to unregister user-mode interface\n"));
    } else {
      dfDebugPrint(1, ("freeDeviceContext: unregistered user-mode interface\n"));
      pDevCtx->userMode.bRegistered = FALSE;
    }
  }

  if (pDevCtx->notification.bTicketPoolValid) {
    dfDebugPrint(1, ("freeDeviceContext: uninitializing notification ticket pool\n"));
    dfPoolUninit(&pDevCtx->notification.ticketPool);
    pDevCtx->notification.bTicketPoolValid = FALSE;
  }

  if (pDevCtx->dma.bTicketPoolValid) {
    dfDebugPrint(1, ("freeDeviceContext: uninitializing DMA ticket pool\n"));
    dfPoolUninit(&pDevCtx->dma.ticketPool);
    pDevCtx->dma.bTicketPoolValid = FALSE;
  }

  if (pDevCtx->clockGenerator.bTicketPoolValid) {
    dfDebugPrint(1, ("freeDeviceContext: uninitializing clock generator ticket pool\n"));
    dfPoolUninit(&pDevCtx->clockGenerator.ticketPool);
    pDevCtx->clockGenerator.bTicketPoolValid = FALSE;
  }
}

static void
stopDevice(
	Admxrc3DeviceContext* pDevCtx)
{
  unsigned int i;
  CoreWindowInfo windowInfo;

  if (NULL != pDevCtx->powerMgmt.hDisableSleep) {
    dfSystemSleepReenable(pDevCtx->pDevObj, pDevCtx->powerMgmt.hDisableSleep);
    pDevCtx->powerMgmt.hDisableSleep = NULL;
    dfDebugPrint(1, ("stopDevice: removed system sleep disable hint\n"));
  }

  if (pDevCtx->userMode.bEnabled) {
    if (!dfInterfaceEnable(pDevCtx->pDevObj, &pDevCtx->userMode.iface, FALSE)) {
      dfDebugPrint(0, ("*** stopDevice: failed to disable user-mode interface\n"));
    } else {
      dfDebugPrint(1, ("stopDevice: disabled user-mode interface\n"));
      pDevCtx->userMode.bEnabled = FALSE;
    }
  }

  if (pDevCtx->bCoreInterfaceImpd) {
    for (i = 0; i < pDevCtx->info.numCommonBuffer; i++) {
      dfMmappableRegionUnregister(pDevCtx->pDevObj, i + ADMXRC3_REGION_TAG_COMMONBUFFER);
      dfDebugPrint(1, ("stopDevice: unregistered mappable common buffer %u", i));
    }

    for (i = 0; i < pDevCtx->info.numWindow; i++) {
      uint64_t base;

      pDevCtx->coreInterface.pGetWindowInfo(pDevCtx->pCoreContext, i, &windowInfo);
      base = (uint64_t)(uintptr_t)windowInfo.pKernelBase;
      dfMmappableRegionUnregister(pDevCtx->pDevObj, i + ADMXRC3_REGION_TAG_WINDOW);
      dfDebugPrint(1, ("stopDevice: unregistered mappable region %u, base=0x%08lx_%08lx\n", i, dfSplitUint64(base)));
    }

    for (i = 0; i < pDevCtx->info.numTargetFpga; i++) {
      if (pDevCtx->fpga[i].alert.bNotificationRegistered) {
        if (pDevCtx->coreInterface.pNotificationUnreg(pDevCtx->pCoreContext, &pDevCtx->fpga[i].alert.notification.coreNotification)) {
          pDevCtx->fpga[i].alert.bNotificationRegistered = FALSE;
          dfDebugPrint(1, ("stopDevice: unregistered for FPGA_ALERT(%u) notification\n", i));
        } else {
          dfDebugPrint(1, ("*** stopDevice: failed to unregister for FPGA_ALERT(%u) notification\n", i));
        }
      }

      if (pDevCtx->fpga[i].interrupt.bNotificationRegistered) {
        if (pDevCtx->coreInterface.pNotificationUnreg(pDevCtx->pCoreContext, &pDevCtx->fpga[i].interrupt.notification.coreNotification)) {
          pDevCtx->fpga[i].interrupt.bNotificationRegistered = FALSE;
          dfDebugPrint(1, ("stopDevice: unregistered for FPGA_INTERRUPT(%u) notification\n", i));
        } else {
          dfDebugPrint(1, ("*** stopDevice: failed to unregister for FPGA_INTERRUPT(%u) notification\n", i));
        }
      }
    }

    if (pDevCtx->sensor.bNotificationRegistered) {
      if (pDevCtx->coreInterface.pNotificationUnreg(pDevCtx->pCoreContext, &pDevCtx->sensor.notification.coreNotification)) {
        pDevCtx->sensor.bNotificationRegistered = FALSE;
        dfDebugPrint(1, ("stopDevice: unregistered for SENSOR_POLL notification\n"));
      } else {
        dfDebugPrint(1, ("*** stopDevice: failed to unregister for SENSOR_POLL notification\n"));
      }
    }

    for (i = 0; i < pDevCtx->info.numFlashBank; i++) {
      if (NULL != pDevCtx->flash[i].pStagingBuffer) {
        dfFree(pDevCtx->flash[i].pStagingBuffer);
        pDevCtx->flash[i].pStagingBuffer = NULL;
        dfDebugPrint(1, ("stopDevice: freed Flash bank %lu staging buffer\n", (unsigned long)i));
      }
    }

    dfReleaseInterface(pDevCtx->pDevObj, g_coreInterfaceName);
    pDevCtx->bCoreInterfaceImpd = FALSE;
    dfDebugPrint(1, ("stopDevice: released core interface\n"));
  }
}

boolean_t
admxrc3OnAdd(
	DfDeviceObject* pDevObj,
	Admxrc3DeviceContext* pDevCtx)
{
  unsigned int i;
  uint32_t val32;

  dfDebugPrint(1, ("admxrc3OnAdd: entered\n"));

  /* Initialize the device context */
	pDevCtx->pDevObj = pDevObj;
  pDevCtx->bCoreInterfaceImpd = FALSE;
  pDevCtx->userMode.bRegistered = FALSE;
  pDevCtx->userMode.bEnabled = FALSE;
  pDevCtx->dma.bTicketPoolValid = FALSE;
  pDevCtx->clockGenerator.bTicketPoolValid = FALSE;
  pDevCtx->notification.bTicketPoolValid = FALSE;
  for (i = 0; i < MAX_NUM_TARGET_FPGA; i++) {
    dfSpinLockInit(&pDevCtx->fpga[i].owner.lock);
    pDevCtx->fpga[i].owner.pClCtx = NULL;
    pDevCtx->fpga[i].interrupt.bNotificationRegistered = FALSE;
#if DF_HAS_USER_EVENT_HANDLE
    dfSpinLockInit(&pDevCtx->fpga[i].interrupt.eventList.lock);
    pDevCtx->fpga[i].interrupt.eventList.pHead = NULL;
    pDevCtx->fpga[i].interrupt.eventList.pTail = NULL;
#endif
    dfSpinLockInit(&pDevCtx->fpga[i].interrupt.waitList.lock);
    dfListInit(&pDevCtx->fpga[i].interrupt.waitList.header);
    pDevCtx->fpga[i].alert.bNotificationRegistered = FALSE;
#if DF_HAS_USER_EVENT_HANDLE
    dfSpinLockInit(&pDevCtx->fpga[i].alert.eventList.lock);
    pDevCtx->fpga[i].alert.eventList.pHead = NULL;
    pDevCtx->fpga[i].alert.eventList.pTail = NULL;
#endif
    dfSpinLockInit(&pDevCtx->fpga[i].alert.waitList.lock);
    dfListInit(&pDevCtx->fpga[i].alert.waitList.header);
    pDevCtx->sensor.bNotificationRegistered = FALSE;
  }
  for (i = 0; i < MAX_NUM_FLASH_BANK; i++) {
    pDevCtx->flash[i].pStagingBuffer = NULL;
  }
  dfSpinLockInit(&pDevCtx->powerMgmt.lock);
  pDevCtx->powerMgmt.openCount = 0;
  pDevCtx->powerMgmt.scheme = PowerMgmtNotActive;
  pDevCtx->powerMgmt.bDeferOpen = FALSE;
  dfListInit(&pDevCtx->powerMgmt.waitingOpen);

  /* From here on, things may fail */

  if (!dfPoolInit(&pDevCtx->clockGenerator.ticketPool, pDevObj, "clockTkt", ClockGeneratorTicket)) {
    dfDebugPrint(0, ("*** admxrc3OnAdd: failed to initialize clock generator ticket pool\n"));
    goto failed;
  }
  pDevCtx->clockGenerator.bTicketPoolValid = TRUE;
  dfDebugPrint(1, ("admxrc3OnAdd: initialized clock generator ticket pool\n"));
  
  if (!dfPoolInit(&pDevCtx->dma.ticketPool, pDevObj, "dmaTkt", DmaTicketUnion)) {
    dfDebugPrint(0, ("*** admxrc3OnAdd: failed to initialize DMA ticket pool\n"));
    goto failed;
  }
  pDevCtx->dma.bTicketPoolValid = TRUE;
  dfDebugPrint(1, ("admxrc3OnAdd: initialized DMA ticket pool\n"));
  
  if (!dfPoolInit(&pDevCtx->notification.ticketPool, pDevObj, "notifyTkt", NotificationTicket)) {
    dfDebugPrint(0, ("*** admxrc3OnAdd: failed to initialize notification ticket pool\n"));
    goto failed;
  }
  pDevCtx->notification.bTicketPoolValid = TRUE;
  dfDebugPrint(1, ("admxrc3OnAdd: initialized notification ticket pool\n"));
  
  if (!dfInterfaceRegister(pDevObj, &pDevCtx->userMode.iface, g_pInterfaceClass, &g_interfaceMethods)) {
    dfDebugPrint(0, ("*** admxrc3OnAdd: failed to register user-mode interface\n"));
    goto failed;
  }
  dfInterfaceSetContext(&pDevCtx->userMode.iface, pDevCtx);
  pDevCtx->userMode.bRegistered = TRUE;
  dfDebugPrint(1, ("admxrc3OnAdd: registered user-mode interface\n"));

  val32 = pDevCtx->powerMgmt.scheme;
  dfParameterGetUint32(pDevCtx->pDevObj->pDrvObj, "PowerManagementScheme", &val32);
  if (val32 < PowerMgmtInvalidValue) {
    pDevCtx->powerMgmt.scheme = (PowerMgmtScheme)val32; /* Cast is safe because of above check */
  }

  return TRUE;

failed:
  freeDeviceContext(pDevCtx);
  return FALSE;
}

void
admxrc3OnRemove(
	DfDeviceObject* pDevObj,
	Admxrc3DeviceContext* pDevCtx)
{
	dfDebugPrint(1, ("admxrc3OnRemove: entered, pDevObj=%p pDevCtx=%p\n", pDevObj, pDevCtx));

	freeDeviceContext(pDevCtx);
}

boolean_t
admxrc3OnStart(
	DfDeviceObject* pDevObj,
  Admxrc3DeviceContext* pDevCtx,
	DfBusResources* pRawRes,
	DfBusResources* pXlatRes)
{
  const size_t maxSizeT = (size_t)-1;
  unsigned int i;
  CoreWindowInfo windowInfo;
  CoreFlashInfo flashInfo;
  CoreFlashStatus flashStatus;

	dfDebugPrint(1, ("admxrc3OnStart: entered, pDevObj=%p pDevCtx=%p\n", pDevObj, pDevCtx));

	if (!dfImportInterface(pDevObj, g_coreInterfaceName, sizeof(Adb3CoreInterface), &pDevCtx->coreInterface, &pDevCtx->pCoreContext)) {
		dfDebugPrint(0, ("*** admxrc3OnStart: failed to import core interface\n"));
		goto failed;
	}
	pDevCtx->bCoreInterfaceImpd = TRUE;
	dfDebugPrint(1, ("admxrc3OnStart: imported core interface\n"));

  pDevCtx->info.model = pDevCtx->coreInterface.pGetModel(pDevCtx->pCoreContext);
  if (!admxrc3Identify(pDevCtx, pDevCtx->info.model)) {
    dfDebugPrint(0, ("*** admxrc3OnStart: unknown model type %lu\n", (unsigned long)pDevCtx->info.model));
    goto failed;
  }

  if (NULL == pDevCtx->methods.pGetBankInfo || NULL == pDevCtx->methods.pGetCardInfo || NULL == pDevCtx->methods.pGetFpgaInfo) {
    dfDebugPrint(0, ("*** admxrc3OnStart: not all methods have been established\n"));
    goto failed;
  }

  pDevCtx->info.numTargetFpga = pDevCtx->coreInterface.pGetNumTargetFpga(pDevCtx->pCoreContext);
  for (i = 0; i < pDevCtx->info.numTargetFpga; i++) {
    pDevCtx->fpga[i].interrupt.notification.targetIndex = i;
    if (!pDevCtx->coreInterface.pNotificationReg(pDevCtx->pCoreContext, CORE_NOTIFY_FPGA_INTERRUPT(i), &pDevCtx->fpga[i].interrupt.notification.coreNotification, &fpgaInterruptNotifyCallback, pDevCtx)) {
      dfDebugPrint(1, ("*** admxrc3OnStart: failed to register for FPGA_INTERRUPT(%u) notification\n", i));
      goto failed;
    } else {
      dfDebugPrint(1, ("admxrc3OnStart: registered for FPGA_INTERRUPT(%u) notification\n", i));
      pDevCtx->fpga[i].interrupt.bNotificationRegistered = TRUE;
    }

    pDevCtx->fpga[i].alert.notification.targetIndex = i;
    if (!pDevCtx->coreInterface.pNotificationReg(pDevCtx->pCoreContext, CORE_NOTIFY_FPGA_ALERT(i), &pDevCtx->fpga[i].alert.notification.coreNotification, &fpgaAlertNotifyCallback, pDevCtx)) {
      dfDebugPrint(1, ("*** admxrc3OnStart: failed to register for FPGA_ALERT(%u) notification\n", i));
      goto failed;
    } else {
      dfDebugPrint(1, ("admxrc3OnStart: registered for FPGA_ALERT(%u) notification\n", i));
      pDevCtx->fpga[i].alert.bNotificationRegistered = TRUE;
    }
  }

#if 0
  if (!pDevCtx->coreInterface.pNotificationReg(pDevCtx->pCoreContext, CORE_NOTIFY_SENSOR_POLL, &pDevCtx->sensor.notification.coreNotification, &sensorPollNotifyCallback, pDevCtx)) {
    dfDebugPrint(1, ("*** admxrc3OnStart: failed to register for SENSOR_POLL notification\n", i));
    goto failed;
  } else {
    dfDebugPrint(1, ("admxrc3OnStart: registered for SENSOR_POLL notification\n", i));
    pDevCtx->sensor.bNotificationRegistered = TRUE;
  }
#endif

  pDevCtx->info.numWindow = pDevCtx->coreInterface.pGetNumWindow(pDevCtx->pCoreContext);
  for (i = 0; i < pDevCtx->info.numWindow; i++) {
    DfMmappableStatus status;

    pDevCtx->coreInterface.pGetWindowInfo(pDevCtx->pCoreContext, i, &windowInfo);
    status = dfMmappableRegionRegister(pDevObj, i + ADMXRC3_REGION_TAG_WINDOW, FALSE, windowInfo.pKernelBase, windowInfo.kernelSize, windowInfo.translatedBase, windowInfo.translatedSize);
    if (DfMmappableSuccess == status) {
      dfDebugPrint(1, ("admxrc3OnStart: registered mappable I/O region, pKernelBase=%p kernelSize=%p translatedBase=0x%08lx_%08lx translatedSize=0x%08lx_%08lx\n",
        windowInfo.pKernelBase, windowInfo.kernelSize, dfSplitUint64(windowInfo.translatedBase), dfSplitUint64(windowInfo.translatedSize)));
    } else {
      dfDebugPrint(0, ("*** admxrc3OnStart: failed to register mappable I/O region, pKernelBase=%p kernelSize=%p translatedBase=0x%08lx_%08lx translatedSize=0x%08lx_%08lx\n",
        windowInfo.pKernelBase, windowInfo.kernelSize, dfSplitUint64(windowInfo.translatedBase), dfSplitUint64(windowInfo.translatedSize)));
      break;
    }
  }
  
  pDevCtx->info.numCommonBuffer = pDevCtx->coreInterface.pGetNumCommonBuffer(pDevCtx->pCoreContext);
  for (i = 0; i < pDevCtx->info.numCommonBuffer; i++) {
    DfMmappableStatus status;
    CoreCommonBufferInfo info;

    pDevCtx->coreInterface.pGetCommonBufferInfo(pDevCtx->pCoreContext, i, &info);
    status = dfMmappableRegionRegister(pDevObj, i + ADMXRC3_REGION_TAG_COMMONBUFFER, TRUE, info.pKernelBase, info.size, 0, 0);
    if (DfMmappableSuccess == status) {
      dfDebugPrint(1, ("admxrc3OnStart: registered mappable common buffer, pKernelBase=%p size=%p\n",
        info.pKernelBase, info.size));
    } else {
      dfDebugPrint(0, ("*** admxrc3OnStart: failed to register mappable common buffer, pKernelBase=%p size=%p\n",
        info.pKernelBase, info.size));
      break;
    }
  }
  
  pDevCtx->info.numFlashBank = pDevCtx->coreInterface.pGetNumFlashBank(pDevCtx->pCoreContext);
  for (i = 0; i < pDevCtx->info.numFlashBank; i++) {
    flashStatus = pDevCtx->coreInterface.pGetFlashInfo(pDevCtx->pCoreContext, 0, &flashInfo);
    if (CoreFlashSuccess == flashStatus) {
      if (flashInfo.largestBlock > maxSizeT) {
        dfDebugPrint(1, ("*** admxrc3OnStart: staging buffer of 0x%08lx_%08lx bytes for Flash bank %lu is too large for this architecture\n",
          dfSplitUint64(flashInfo.largestBlock), (unsigned long)i));
        goto failed;
      }
      pDevCtx->flash[i].pStagingBuffer = (uint8_t*)dfMalloc((size_t)flashInfo.largestBlock); /* Cast safe due to check immediately above */
      if (NULL == pDevCtx->flash[i].pStagingBuffer) {
        dfDebugPrint(1, ("*** admxrc3OnStart: failed to allocate 0x%08lx_%08lx bytes for Flash bank %lu staging buffer\n",
          dfSplitUint64(flashInfo.largestBlock), (unsigned long)i));
        goto failed;
      }
    }
  }
  
  if (!dfInterfaceEnable(pDevObj, &pDevCtx->userMode.iface, TRUE)) {
    dfDebugPrint(0, ("*** admxrc3OnStart: failed to enable user-mode interface\n"));
    goto failed;
  } else {
    dfDebugPrint(1, ("admxrc3OnStart: enabled user-mode interface\n"));
    pDevCtx->userMode.bEnabled = TRUE;
  }

  if (pDevCtx->powerMgmt.scheme == PowerMgmtNone) {
    pDevCtx->powerMgmt.hDisableSleep = dfSystemSleepDisable(pDevCtx->pDevObj);
    if (NULL == pDevCtx->powerMgmt.hDisableSleep) {
      dfDebugPrint(0, ("*** admxrc3OnStart: failed to add system sleep disable hint\n"));
      goto failed;
    }
  } else {
    dfDebugPrint(1, ("admxrc3OnStart: power management scheme is 'none', added system sleep disable hint\n"));
  }

  return TRUE;

failed:
	stopDevice(pDevCtx);
	return FALSE;
}

void
admxrc3OnStop(
	DfDeviceObject* pDevObj,
	Admxrc3DeviceContext* pDevCtx)
{
	dfDebugPrint(1, ("admxrc3OnStop: entered, pDevObj=%p pDevCtx=%p\n", pDevObj, pDevCtx));

	stopDevice(pDevCtx);
}

boolean_t
admxrc3OnPowerQuery(
	DfDeviceObject* pDevObj,
	Admxrc3DeviceContext* pDevCtx,
  unsigned int currentState,
  unsigned int requestedState,
  boolean_t bIsShutdown)
{
  DfSpinLockFlags f;
  boolean_t bVeto = FALSE;

  dfDebugPrint(1, ("admxrc3OnPowerQuery: entered, pDevObj=%p pDevCtx=%p currentState=D%u requestedState=D%u bIsShutdown=%s\n",
    pDevObj, pDevCtx, currentState, requestedState, bIsShutdown ? "TRUE" : "FALSE"));

  dfAssert(requestedState != currentState);

  if (requestedState <= currentState) {
    /* Wants to power up - always allow */
    dfAssert(!bIsShutdown);
    return FALSE;
  }

  if (bIsShutdown) {
    /* If shutting down, don't bother to veto power change */
    return FALSE;
  }

  /* Powering down */
  f = dfSpinLockGet(&pDevCtx->powerMgmt.lock);
  switch (pDevCtx->powerMgmt.scheme) {
  case PowerMgmtNotActive:
    if (pDevCtx->powerMgmt.openCount) {
      bVeto = TRUE; /* Veto power state change because device is open somewhere */
    } else {
      /* Start refusing open attempts */
      pDevCtx->powerMgmt.bDeferOpen = TRUE;
    }
    break;

  case PowerMgmtNone:
  default:
    bVeto = TRUE; /* Veto power state change */
    break;
  }
  dfSpinLockPut(&pDevCtx->powerMgmt.lock, f);

  return bVeto;
}

static void
completeWaitingOpens(
  Admxrc3DeviceContext* pDevCtx,
  DfIoStatus status)
{
  DfListNode* p;
  Admxrc3ClientContext* pClCtx;
  unsigned int count = 0;

  dfDebugPrint(1, ("completeWaitingOpens: completing opens\n"));

  while (!dfListIsEmpty(&pDevCtx->powerMgmt.waitingOpen)) {
    p = dfListGetHead(&pDevCtx->powerMgmt.waitingOpen);
    dfListRemove(p);
    pClCtx = DF_CONTAINER_OF(p, Admxrc3ClientContext, waitNode);
    dfRequestComplete(pClCtx->pOpenRequest, status);
    count++;
  }

  dfDebugPrint(1, ("completeWaitingOpens: completed %u open(s)\n", count));
}

void
admxrc3OnPowerSet(
	DfDeviceObject* pDevObj,
	Admxrc3DeviceContext* pDevCtx,
  unsigned int currentState,
  unsigned int newState,
  boolean_t bIsShutdown)
{
  DfSpinLockFlags f;

  dfDebugPrint(1, ("admxrc3OnPowerSet: entered, pDevObj=%p pDevCtx=%p currentState=D%u newState=D%u bIsShutdown=%s\n",
    pDevObj, pDevCtx, currentState, newState, bIsShutdown ? "TRUE" : "FALSE"));

  if (newState <= currentState) {
    /* Powering up or reaffirming current working state */
    dfAssert(!bIsShutdown);
    f = dfSpinLockGet(&pDevCtx->powerMgmt.lock);
    switch (pDevCtx->powerMgmt.scheme) {
    case PowerMgmtNotActive:
      /* Stop deferring open attempts */
      pDevCtx->powerMgmt.bDeferOpen = FALSE;
      dfSpinLockPut(&pDevCtx->powerMgmt.lock, f);
      /* Complete all waiting open attempts */
      completeWaitingOpens(pDevCtx, DfIoStatusSuccess);
      break;

    case PowerMgmtNone:
    default:
      dfSpinLockPut(&pDevCtx->powerMgmt.lock, f);
      /* Nothing to do */
      break;
    }
  }
}

boolean_t
admxrc3DriverEntry(
  DfDriverObject* pDrvObj)
{
  dfDebugPrint(1, ("admxrc3DriverEntry: entered\n"));

#if defined(ADMXRC3_LOCKED_BUFFERS_GLOBAL) && ADMXRC3_LOCKED_BUFFERS_GLOBAL
  initLockedBuffers(pDrvObj);
#endif

  g_interfaceMethods.pOnCleanup = onCleanup;
  g_interfaceMethods.pOnClose = onClose;
  g_interfaceMethods.pOnIoctl = onIoctl;
#if DF_NEED_THUNK
  g_interfaceMethods.pOnIoctlThunk = onIoctlThunk;
#endif
  g_interfaceMethods.pOnOpen = onOpen;

  if (!dfInterfaceClassRegister(pDrvObj, "admxrc3", &g_pInterfaceClass)) {
    dfDebugPrint(0, ("*** admxrc3DriverEntry: failed to register interface class '%s'\n", g_interfaceClassName));
    return FALSE;
  }
  dfDebugPrint(1, ("admxrc3DriverEntry: registered interface class '%s', pClass=%p\n", g_interfaceClassName, g_pInterfaceClass));

  return TRUE;
}

extern void
admxrc3DriverExit(
  DfDriverObject* pDrvObj)
{ 
  dfDebugPrint(1, ("admxrc3DriverExit: entered\n"));

  if (NULL != g_pInterfaceClass) {
    dfDebugPrint(1, ("admxrc3DriverExit: unregistering interface class '%s', pClass=%p\n", g_interfaceClassName, g_pInterfaceClass));
    dfInterfaceClassUnregister(g_pInterfaceClass);
    g_pInterfaceClass = NULL;
  }
}

/*
** ---------------------------------------------------------------------------------
** These functions are built into the modular driver, but not the monolithic driver
** ---------------------------------------------------------------------------------
*/

#if !ADB3_MONOLITHIC
static boolean_t
onAdd(
	DfDeviceObject* pDevObj)
{
	Admxrc3DeviceContext* pDevCtx;

  dfDebugPrint(1, ("onAdd: entered, pDevObj=%p\n", pDevObj));

  pDevCtx = (Admxrc3DeviceContext*)dfDeviceGetContext(pDevObj);
	pDevCtx->pDevObj = pDevObj;

  return admxrc3OnAdd(pDevObj, pDevCtx);
}
#endif

#if !ADB3_MONOLITHIC
static void
onRemove(
	DfDeviceObject* pDevObj)
{
	Admxrc3DeviceContext* pDevCtx;

	dfDebugPrint(1, ("onRemove: entered, pDevObj=%p\n", pDevObj));

	pDevCtx = (Admxrc3DeviceContext*)dfDeviceGetContext(pDevObj);
  admxrc3OnRemove(pDevObj, pDevCtx);
}
#endif

#if !ADB3_MONOLITHIC
static boolean_t
onStart(
	DfDeviceObject* pDevObj,
	DfBusResources* pRawRes,
	DfBusResources* pXlatRes)
{
  Admxrc3DeviceContext* pDevCtx;

	dfDebugPrint(1, ("onStart: entered, pDevObj=%p\n", pDevObj));

	pDevCtx = (Admxrc3DeviceContext*)dfDeviceGetContext(pDevObj);
  return admxrc3OnStart(pDevObj, pDevCtx, pRawRes, pXlatRes);
}
#endif

#if !ADB3_MONOLITHIC
static void
onStop(
	DfDeviceObject* pDevObj)
{
	Admxrc3DeviceContext* pDevCtx;

	dfDebugPrint(1, ("onStop: entered, pDevObj=%p\n", pDevObj));

	pDevCtx = (Admxrc3DeviceContext*)dfDeviceGetContext(pDevObj);
  admxrc3OnStop(pDevObj, pDevCtx);
}
#endif

#if !ADB3_MONOLITHIC
static boolean_t
onPowerQuery(
	DfDeviceObject* pDevObj,
  unsigned int currentState,
  unsigned int requestedState,
  boolean_t bIsShutdown)
{
	Admxrc3DeviceContext* pDevCtx;

	pDevCtx = (Admxrc3DeviceContext*)dfDeviceGetContext(pDevObj);

  dfDebugPrint(1, ("onPowerQuery: entered, pDevObj=%p pDevCtx=%p currentState=D%u requestedState=D%u bIsShutdown=%s\n",
    pDevObj, pDevCtx, currentState, requestedState, bIsShutdown ? "TRUE" : "FALSE"));

  return admxrc3OnPowerQuery(pDevObj, pDevCtx, currentState, requestedState, bIsShutdown);
}

static void
onPowerSet(
	DfDeviceObject* pDevObj,
  unsigned int currentState,
  unsigned int newState,
  boolean_t bIsShutdown)
{
	Admxrc3DeviceContext* pDevCtx;

	pDevCtx = (Admxrc3DeviceContext*)dfDeviceGetContext(pDevObj);

  dfDebugPrint(1, ("onPowerSet: entered, pDevObj=%p pDevCtx=%p currentState=D%u newState=D%u  bIsShutdown=%s\n",
    pDevObj, pDevCtx, currentState, newState, bIsShutdown ? "TRUE" : "FALSE"));

  admxrc3OnPowerSet(pDevObj, pDevCtx, currentState, newState, bIsShutdown);
}
#endif

#if !ADB3_MONOLITHIC
boolean_t
dfDriverEntry(
	DfDriverObject* pDrvObj)
{
	dfDebugPrint(1, ("dfDriverEntry: ADMXRC3 Module Driver, pDrvObj=%p version=%lu.%lu.%lu.%lu\n",
    pDrvObj,
    (unsigned long)ADB3_VERSION_0, (unsigned long)ADB3_VERSION_1, (unsigned long)ADB3_VERSION_2, (unsigned long)ADB3_VERSION_3));

  pDrvObj->deviceContextSize = sizeof(Admxrc3DeviceContext);
  pDrvObj->methods.pOnAdd = onAdd;
  pDrvObj->methods.pOnRemove = onRemove;
  pDrvObj->methods.pOnStart = onStart;
  pDrvObj->methods.pOnStop = onStop;
  pDrvObj->methods.pOnPowerQuery = onPowerQuery;
  pDrvObj->methods.pOnPowerSet = onPowerSet;

  return admxrc3DriverEntry(pDrvObj);
}
#endif

#if !ADB3_MONOLITHIC
extern void
dfDriverExit(
	DfDriverObject* pDrvObj)
{
	dfDebugPrint(1, ("dfDriverExit: ADMXRC3 Module Driver exiting\n"));

  admxrc3DriverExit(pDrvObj);
}
#endif
