/*
** File: model_admpcie7v3.c
** Project: ADMXRC3 module driver
** Purpose: ADM-PCIE-7V3 specific code, including the function to bootstrap
**          the device context.
**
** (C) Copyright Alpha Data 2014
*/

#include <df.h>

#include "device.h"
#include "info.h"
#include "model_boot.h"
#include <core/admpcie7v3.h>

static ADMXRC3_STATUS
getBankInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int bankIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETBANKINFO* pIoctl)
{
  VpdAdmpcie7v3* pVpd = (VpdAdmpcie7v3*)pVpdBuffer;
  uint64_t physicalSize;

  if (bankIndex >= DF_ARRAY_LENGTH(pVpd->dimm)) {
    return ADMXRC3_INVALID_INDEX;
  }
  physicalSize = pVpd->dimm[bankIndex].physicalSize;
  pIoctl->out.typeMask = ADMXRC3_BANK_SDRAM_DDR3;
  pIoctl->out.physicalDataWidth = 64U;
  pIoctl->out.physicalEccWidth = (uint16_t)(pVpd->dimm[bankIndex].bHasEcc ? 8U : 0U);
  pIoctl->out.maximumFrequency = (uint64_t)pVpd->dimm[bankIndex].freqMax;
  pIoctl->out.minimumFrequency = 303030303U;
  pIoctl->out.connectivityMask = 0x1U;
  pIoctl->out.physicalSize = physicalSize;
  pIoctl->out.bPresent = (uint8_t)(physicalSize != 0);

  return ADMXRC3_SUCCESS;
}

static void
getCardInfo(
	Admxrc3DeviceContext* pDevCtx,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETCARDINFO* pIoctl)
{
  VpdAdmpcie7v3* pVpd = (VpdAdmpcie7v3*)pVpdBuffer;
  unsigned int i;

  pIoctl->out.serialNumber = pVpd->prom.rev0.serial;
  pIoctl->out.numMemoryBank = DF_ARRAY_LENGTH(pVpd->dimm);
  pIoctl->out.memoryBankPresent = 0;
  for (i = 0; i < DF_ARRAY_LENGTH(pVpd->dimm); i++) {
    if (0 != pVpd->dimm[i].physicalSize) {
      /* Bank is fitted */
      pIoctl->out.memoryBankPresent |= (0x1U << i);
    }
  }
}

static ADMXRC3_STATUS
getFpgaInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int targetIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETFPGAINFO* pIoctl)
{
  VpdAdmpcie7v3* pVpd = (VpdAdmpcie7v3*)pVpdBuffer;
  ADMXRC3_PACKAGE_TYPE package;
  ADMXRC3_FPGA_TYPE fpgaType;
  char speed[8];
  char stepping[8];

  switch (targetIndex) {
  case 0:
    fpgaType = (ADMXRC3_FPGA_TYPE)pVpd->prom.rev0.fpga.device;
    break;

  default:
    return ADMXRC3_INVALID_INDEX;
  }

  package = ADMXRC3_PACKAGE_MAKEFF(1157);

  dfAssert(sizeof(char) == sizeof(uint8_t));

  dfStrCpy(speed, DF_ARRAY_LENGTH(speed), (const char*)pVpd->prom.rev0.fpga.speedGrade);
  switch (pVpd->prom.rev0.fpga.tempGrade) {
  case ADMPCIE7V3_VPD_TEMPGRADE_COMMERCIAL:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "C");
    break;

  case ADMPCIE7V3_VPD_TEMPGRADE_INDUSTRIAL:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "I");
    break;

  case ADMPCIE7V3_VPD_TEMPGRADE_EXTENDED:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "E");
    break;

  default:
    dfStrCat(speed, DF_ARRAY_LENGTH(speed), "?");
    break;
  }

  dfStrCpy(stepping, DF_ARRAY_LENGTH(stepping), (const char*)pVpd->prom.rev0.fpga.stepping);
  pIoctl->out.flags = ADMXRC3_FPGA_SPEEDVALID | ADMXRC3_FPGA_NOTCONFIGURABLE;
  if (stepping[0] != '\0') {
    pIoctl->out.flags |= ADMXRC3_FPGA_STEPPINGVALID;
  }
  setupIoctlsGetFpgaInfo(pDevCtx, pIoctl, fpgaType, package, speed, stepping, TRUE);

  return ADMXRC3_SUCCESS;
}

void
admxrc3BootstrapAdmpcie7v3(
	Admxrc3DeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("admxrc3BootstrapAdmpcie7v3: entered\n"));

  pDevCtx->methods.pGetBankInfo = getBankInfo;
  pDevCtx->methods.pGetCardInfo = getCardInfo;
  pDevCtx->methods.pGetFpgaInfo = getFpgaInfo;
  pDevCtx->bReverseSelectMapData = FALSE;
  pDevCtx->localAddressLimit = (uint64_t)(0x1ULL << 39U);
  pDevCtx->localWidth = 16U;
}
